/*
 * Copyright (c) 2010 Tomasz Krasuski (see AUTHORS)
 *
 * This file is part of the A8CAS project which allows to manipulate tape
 * images for Atari 8-bit computers.
 *
 * A8CAS is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * A8CAS is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with A8CAS; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 */
#include <assert.h>

#include "cas_decode.h"

#include "a8cas_file.h"
#include "a8cas.h"
#include "cas_header.h"

enum { DEFAULT_BAUDRATE = 600 };

static int read_recognise_block(CAS_DECODE_t *decode, A8CAS_signal *sig);
static int read_data_block_next_byte(CAS_DECODE_t *decode, A8CAS_signal *sig);

static int read_data_block_next_bit(CAS_DECODE_t *decode, A8CAS_signal *sig)
{
	char current;

	sig->length = 0;

	sig->signal = (decode->current_10bit & decode->bit_mask) != 0;
	while (decode->bit_mask != 0x400
	       && (current = (decode->current_10bit & decode->bit_mask) != 0) == sig->signal) {
		sig->length ++;
		(decode->bit_mask) <<= 1;
	};
	sig->rate = decode->baudrate;
	if (decode->bit_mask == 0x400)
		/* End of byte */
		decode->read_func = &read_data_block_next_byte;
	return 0;
}

/* Invoked before processing the next byte from the buffer, this function
 * splits the byte on 10 bits and fills the current_10bit value. Then the
 * bit processing starts.
 * Returns 0 on success, otherwise 1 and sets file->info.errnum.
 */
static int read_data_block_next_byte(CAS_DECODE_t *decode, A8CAS_signal *sig)
{
	uint8_t byte;
	int retval = (*decode->read_data_byte_func)(decode->file, &byte);

	if (retval == 2) {
		/* End of DATA block */
		decode->read_func = &read_recognise_block;
		(*decode->next_block_func)(decode->file);
		return read_recognise_block(decode, sig);
	}
	else if (retval != 0)
		return 1;
	decode->current_10bit = ((unsigned int)byte << 1) | 0x200;
	decode->bit_mask = 1;
	decode->read_func = &read_data_block_next_bit;
	return read_data_block_next_bit(decode, sig);
}

static int read_fsk_block_next_signal(CAS_DECODE_t *decode, A8CAS_signal *sig)
{
	uint16_t length;
	int retval;

	sig->signal = (char)decode->bit_mask;
	decode->bit_mask = ! decode->bit_mask;
	retval = (*decode->read_fsk_signal_func)(decode->file, &length);
	if (retval == 2) {
		/* End of FSK block */
		decode->read_func = &read_recognise_block;
		(*decode->next_block_func)(decode->file);
		return read_recognise_block(decode, sig);
	}
	else if (retval != 0)
		return 1;
	sig->length = (unsigned int) length;
	sig->rate = 10000;
	return 0;
}

/* Invoked just after reading a new CAS block, this function sets FSK to
 * the signal of the new block's IRG gap.
 */
static void read_irg(CAS_DECODE_t *decode, A8CAS_signal *sig, CAS_HEADER_t *header)
{
	sig->signal = 1;
	sig->rate = 1000;
	sig->length = header->aux;
}

/* The "entry" point for all CAS read functions. Its purpose is to read the
   next block's header, recognise its type and redirect the cas_decode->read_func
   pointer so that next calls to read functions will perform block-specific
   actions.
   Returns 0 on success; otherwise 1 and sets file->info.errnum. */
static int read_recognise_block(CAS_DECODE_t *decode, A8CAS_signal *sig)
{
	for (;;) {
		CAS_HEADER_t header;
		if ((*decode->read_header_func)(decode->file, &header) != 0)
			return 1;

		/* Remember not to use the header.length field - it's not initialised. */
		switch (header.type) {
		case BLOCK_TYPE_BAUD:
			if (header.aux == 0) {
				/* Baudrate of 0 is not allowed (divide by 0!). */
				decode->file->info.errnum = A8CAS_ERR_BADFILE;
				return 1;
			}
			decode->baudrate = header.aux;
			break;
		case BLOCK_TYPE_DATA:
			decode->read_func = &read_data_block_next_byte;
			read_irg(decode, sig, &header);
			return 0;
		case BLOCK_TYPE_FSK:
			decode->read_func = &read_fsk_block_next_signal;
			decode->bit_mask = 0;
			read_irg(decode, sig, &header);
			return 0;
		default:
			decode->file->info.errnum = A8CAS_ERR_BADFILE;
			return 1;
		}
	}
	return 0;
}

void CAS_DECODE_reset(CAS_DECODE_t *decode, uint16_t baudrate)
{
	decode->read_func = &read_recognise_block;
	decode->baudrate = (baudrate == 0 ? DEFAULT_BAUDRATE : baudrate);
	decode->last_irg_signal.length = 0;
}

int CAS_DECODE_read(CAS_DECODE_t *decode, A8CAS_signal *sig)
{
	return (*decode->read_func)(decode, sig);
}

/* ------------------------------------------------------------------------ *
 * Fuunctions to read full bytes/signals                                    *
 * ------------------------------------------------------------------------ */

unsigned int CAS_DECODE_read_bytes(CAS_DECODE_t *decode, unsigned char *bytes, unsigned int size, unsigned int *baudrate)
{
	unsigned int written = 0;

	for (;;) {

		if (written >= size) {
			/* Result buffer full. */
			decode->file->info.errnum = A8CAS_ERR_NONE;
			return written;
		}

		if (decode->last_irg_signal.length != 0) {
			decode->file->info.errnum = A8CAS_ERR_NONE;
			return written;
		}
		if (decode->read_func == &read_recognise_block) {
			if (read_recognise_block(decode, &decode->last_irg_signal) != 0) {
				if (decode->file->info.errnum == A8CAS_ERR_EOF)
					return written;
				else
					return 0;
			}
			continue;
		}
		if (decode->read_func == &read_data_block_next_bit
		|| decode->read_func == &read_fsk_block_next_signal) {
			/* Reading a signal; no full byte to read. */
			decode->file->info.errnum = A8CAS_ERR_NONE;
			return written;
		}

		assert(decode->read_func == &read_data_block_next_byte);

		*baudrate = (unsigned int) decode->baudrate;
		for ( ; written < size; written ++) {
			int retval = (*decode->read_data_byte_func)(decode->file, &bytes[written]);
			if (retval == 2) {
				/* End of data block */
				decode->read_func = &read_recognise_block;
				(*decode->next_block_func)(decode->file);
				break;
			}
			else if (retval != 0)
				/* Some error. */
				return 0;
		}
	}
}

unsigned int CAS_DECODE_read_signals(CAS_DECODE_t *decode, A8CAS_signal *sigs, unsigned int size)
{
	unsigned int written = 0;

	for (;;) {
		if (written >= size) {
			decode->file->info.errnum = A8CAS_ERR_NONE;
			return written;
		}

		if (decode->last_irg_signal.length != 0) {
			sigs[written ++] = decode->last_irg_signal;
			decode->last_irg_signal.length = 0;
			continue;
		}
		if (decode->read_func == &read_recognise_block) {
			if (read_recognise_block(decode, &decode->last_irg_signal) != 0) {
				if (decode->file->info.errnum == A8CAS_ERR_EOF)
					return written;
				else
					return 0;
			}
			continue;
		}

		if (decode->read_func == &read_data_block_next_byte) {
			/* There's a byte to read. */
			decode->file->info.errnum = A8CAS_ERR_NONE;
			return written;
		}
		else if (decode->read_func == &read_data_block_next_bit) {
			do {
				read_data_block_next_bit(decode, &sigs[written++]);
			} while (decode->read_func == &read_data_block_next_bit);
			continue;
		}

		assert(decode->read_func == &read_fsk_block_next_signal);

		for ( ; written < size; written ++) {
			uint16_t length;
			int retval = (*decode->read_fsk_signal_func)(decode->file, &length);
			if (retval == 2) {
				decode->read_func = &read_recognise_block;
				(*decode->next_block_func)(decode->file);
				break;
			}
			else if (retval != 0)
				/* Some error. */
				return 0;
			sigs[written].length = length;
			sigs[written].rate = 10000;
			sigs[written].signal = (char)decode->bit_mask;
			decode->bit_mask = ! decode->bit_mask;
		}
	}
}
